// Copyright (c) 2019-2022, Lawrence Livermore National Security, LLC and
// other Serac Project Developers. See the top-level LICENSE file for
// details.
//
// SPDX-License-Identifier: (BSD-3-Clause)

#include "axom/slic/core/SimpleLogger.hpp"
#include <gtest/gtest.h>

#include "serac/numerics/expr_template_ops.hpp"

static std::pair<mfem::Vector, mfem::Vector> sample_vectors(const int entries)
{
  mfem::Vector lhs(entries);
  mfem::Vector rhs(entries);
  for (int i = 0; i < entries; i++) {
    lhs[i] = i * 4 + 1;
    rhs[i] = i * i * 3 + 2;
  }
  return {lhs, rhs};
}

static std::pair<mfem::DenseMatrix, mfem::Vector> sample_matvec(const int rows, const int cols)
{
  mfem::Vector      vec_in(cols);
  mfem::DenseMatrix matrix(rows, cols);
  for (int i = 0; i < cols; i++) {
    vec_in[i] = i * 4 + 1;
    for (int j = 0; j < rows; j++) {
      matrix(j, i) = 2 * (i == j) - (i == (j + 1)) - (i == (j - 1));
    }
  }
  return {matrix, vec_in};
}

static auto build_partitioning(MPI_Comm comm, const int size)
{
  int num_procs = 0;
  int rank      = 0;
  MPI_Comm_size(comm, &num_procs);
  MPI_Comm_rank(comm, &rank);
  bool      assumed_partition = HYPRE_AssumedPartitionCheck();
  auto      partitioning = std::make_unique<int[]>(assumed_partition ? 2 : static_cast<std::size_t>(num_procs + 1));
  const int per_proc     = (size / num_procs) + ((size % num_procs != 0) ? 1 : 0);

  if (assumed_partition) {
    auto n_entries  = (rank == num_procs - 1) ? size - ((num_procs - 1) * per_proc) : per_proc;
    partitioning[0] = per_proc * rank;
    partitioning[1] = (per_proc * rank) + n_entries;
  } else {
    for (int i = 0; i < num_procs; i++) {
      partitioning[static_cast<std::size_t>(i)] = per_proc * i;
    }
    partitioning[static_cast<std::size_t>(num_procs)] = size;
  }
  return std::make_pair(std::move(partitioning), per_proc * rank);
}

TEST(ExprTemplates, BasicAdd)
{
  MPI_Barrier(MPI_COMM_WORLD);
  constexpr int size = 10;
  auto [lhs, rhs]    = sample_vectors(size);

  mfem::Vector mfem_result(size);
  add(lhs, rhs, mfem_result);

  mfem::Vector expr_result = lhs + rhs;

  for (int i = 0; i < size; i++) {
    EXPECT_DOUBLE_EQ(mfem_result[i], expr_result[i]);
  }
  MPI_Barrier(MPI_COMM_WORLD);
}

TEST(ExprTemplates, BasicAddHyprepar)
{
  MPI_Barrier(MPI_COMM_WORLD);
  constexpr int size = 10;
  auto [lhs, rhs]    = sample_vectors(size);

  auto [partitioning, start] = build_partitioning(MPI_COMM_WORLD, size);

  mfem::HypreParVector lhs_par(MPI_COMM_WORLD, size, lhs + start, partitioning.get());
  mfem::HypreParVector rhs_par(MPI_COMM_WORLD, size, rhs + start, partitioning.get());

  mfem::HypreParVector mfem_result(MPI_COMM_WORLD, size, partitioning.get());
  add(lhs_par, rhs_par, mfem_result);

  mfem::HypreParVector expr_result(MPI_COMM_WORLD, size, partitioning.get());
  evaluate(lhs_par + rhs_par, expr_result);

  double* mfem_local = mfem_result;
  double* expr_local = expr_result;

  for (int i = 0; i < expr_result.Size(); i++) {
    EXPECT_DOUBLE_EQ(mfem_local[i], expr_local[i]);
  }

  MPI_Barrier(MPI_COMM_WORLD);
}

TEST(ExprTemplates, BasicDiv)
{
  MPI_Barrier(MPI_COMM_WORLD);
  constexpr int    size   = 10;
  constexpr double scalar = 0.3;
  mfem::Vector     a(size);
  mfem::Vector     mfem_result(size);
  for (int i = 0; i < size; i++) {
    a[i]           = i * 4 + 1;
    mfem_result[i] = a[i];
  }

  // Dividing a vector by a scalar
  mfem_result /= scalar;

  mfem::Vector expr_result = a / scalar;

  for (int i = 0; i < size; i++) {
    EXPECT_DOUBLE_EQ(mfem_result[i], expr_result[i]);
  }

  // Dividing a scalar by a vector
  mfem_result = a;
  for (int i = 0; i < size; i++) {
    a[i] = scalar / a[i];
  }

  expr_result = scalar / a;

  for (int i = 0; i < size; i++) {
    EXPECT_DOUBLE_EQ(mfem_result[i], expr_result[i]);
  }
  MPI_Barrier(MPI_COMM_WORLD);
}

TEST(ExprTemplates, BasicAddLambda)
{
  MPI_Barrier(MPI_COMM_WORLD);
  constexpr int size = 10;
  auto [lhs, rhs]    = sample_vectors(size);

  mfem::Vector mfem_result(size);
  add(lhs, rhs, mfem_result);

  auto lambda_add = [](const auto& l, const auto& r) { return l + r; };

  mfem::Vector expr_result = lambda_add(lhs, rhs);

  EXPECT_EQ(mfem_result.Size(), expr_result.Size());
  for (int i = 0; i < size; i++) {
    EXPECT_DOUBLE_EQ(mfem_result[i], expr_result[i]);
  }
  MPI_Barrier(MPI_COMM_WORLD);
}

TEST(ExprTemplates, SubtractionNotCommutative)
{
  MPI_Barrier(MPI_COMM_WORLD);
  constexpr int size = 10;
  mfem::Vector  a(size);
  mfem::Vector  b(size);
  mfem::Vector  c(size);

  for (int i = 0; i < size; i++) {
    a[i] = i * 4 + 1;
    b[i] = i * i * 3 + 2;
    c[i] = i * i * i * 7 + 23;
  }

  // Tests that switching the order of operations
  // does not change the result
  mfem::Vector result1 = a + c - b;  // Parsed as (a + c) - b
  mfem::Vector result2 = c - b + a;  // Parsed as (c - b) + a
  for (int i = 0; i < size; i++) {
    EXPECT_DOUBLE_EQ(result1[i], result2[i]);
  }

  mfem::Vector result3 = a - c;
  mfem::Vector result4 = c - a;
  for (int i = 0; i < size; i++) {
    EXPECT_FALSE(result3[i] == result4[i]);
  }
  MPI_Barrier(MPI_COMM_WORLD);
}

TEST(ExprTemplates, SubtractionNotCommutativeRvalue)
{
  MPI_Barrier(MPI_COMM_WORLD);
  constexpr int size           = 3;
  double        a_values[size] = {-12.2692, 6.23918, -12.2692};
  double        b_values[size] = {0.0850848, -0.17017, 0.0850848};

  auto fext = [](const double t) {
    mfem::Vector force(3);
    force[0] = -10 * t;
    force[1] = 0;
    force[2] = 10 * t * t;
    return force;
  };

  mfem::Vector a(a_values, size);
  mfem::Vector b(b_values, size);
  mfem::Vector c        = fext(0.0);
  mfem::Vector resulta1 = c - a - b;
  mfem::Vector resulta2 = fext(0.0) - a - b;
  for (int i = 0; i < size; i++) {
    EXPECT_DOUBLE_EQ(resulta1[i], resulta2[i]);
  }
  MPI_Barrier(MPI_COMM_WORLD);
}

TEST(ExprTemplates, AdditionCommutative)
{
  MPI_Barrier(MPI_COMM_WORLD);
  constexpr int size = 10;
  auto [a, b]        = sample_vectors(size);

  mfem::Vector result1 = a + b;
  mfem::Vector result2 = b + a;
  for (int i = 0; i < size; i++) {
    EXPECT_DOUBLE_EQ(result1[i], result2[i]);
  }
  MPI_Barrier(MPI_COMM_WORLD);
}

TEST(ExprTemplates, ScalarMultCommutative)
{
  MPI_Barrier(MPI_COMM_WORLD);
  constexpr int size = 10;
  mfem::Vector  a(size);
  double        scalar = 0.3;

  for (int i = 0; i < size; i++) {
    a[i] = i * 4 + 1;
  }

  mfem::Vector result1 = scalar * a;
  mfem::Vector result2 = a * scalar;
  for (int i = 0; i < size; i++) {
    EXPECT_DOUBLE_EQ(result1[i], result2[i]);
  }
  MPI_Barrier(MPI_COMM_WORLD);
}

TEST(ExprTemplates, MoveFromTempLambda)
{
  MPI_Barrier(MPI_COMM_WORLD);
  constexpr int size = 10;
  auto [lhs, rhs]    = sample_vectors(size);

  mfem::Vector mfem_result(size);
  add(lhs, 3.5, rhs, mfem_result);

  auto lambda_add = [](const auto& l, const auto& r) {
    auto r35 = r * 3.5;
    return l + std::move(r35);
  };

  mfem::Vector expr_result = lambda_add(lhs, rhs);

  EXPECT_EQ(mfem_result.Size(), expr_result.Size());
  for (int i = 0; i < size; i++) {
    EXPECT_DOUBLE_EQ(mfem_result[i], expr_result[i]);
  }
  MPI_Barrier(MPI_COMM_WORLD);
}

TEST(ExprTemplates, MoveFromTempVecLambda)
{
  MPI_Barrier(MPI_COMM_WORLD);
  constexpr int size = 10;
  auto [lhs, rhs]    = sample_vectors(size);

  mfem::Vector mfem_result(size);
  add(lhs, 3.5, rhs, mfem_result);

  auto lambda_add = [](const auto& l, const auto& r) {
    mfem::Vector r35 = r * 3.5;
    return l + std::move(r35);
  };

  mfem::Vector expr_result = lambda_add(lhs, rhs);

  EXPECT_EQ(mfem_result.Size(), expr_result.Size());
  for (int i = 0; i < size; i++) {
    EXPECT_DOUBLE_EQ(mfem_result[i], expr_result[i]);
  }
  MPI_Barrier(MPI_COMM_WORLD);
}

TEST(ExprTemplates, SmallMatvec)
{
  MPI_Barrier(MPI_COMM_WORLD);
  constexpr int rows    = 10;
  constexpr int cols    = 12;
  auto [matrix, vec_in] = sample_matvec(rows, cols);

  mfem::Vector mfem_result(rows);
  matrix.Mult(vec_in, mfem_result);

  mfem::Vector expr_result = matrix * vec_in;

  EXPECT_EQ(mfem_result.Size(), expr_result.Size());
  for (int i = 0; i < rows; i++) {
    EXPECT_DOUBLE_EQ(mfem_result[i], expr_result[i]);
  }
  MPI_Barrier(MPI_COMM_WORLD);
}

TEST(ExprTemplates, SmallMixedExpr)
{
  MPI_Barrier(MPI_COMM_WORLD);
  constexpr int rows = 10;
  auto [lhs, rhs]    = sample_vectors(rows);

  constexpr int cols    = 12;
  auto [matrix, vec_in] = sample_matvec(rows, cols);

  mfem::Vector mfem_result(rows);
  mfem::Vector expr_result(rows);

  // -lhs + rhs * 3.0 - 0.3 * (matrix * vec_in)
  mfem::Vector matvec(rows);
  matrix.Mult(vec_in, matvec);

  mfem::Vector vec_negate_scale(rows);
  add(-1.0, lhs, 3.0, rhs, vec_negate_scale);

  add(vec_negate_scale, -0.3, matvec, mfem_result);

  expr_result = -lhs + rhs * 3.0 - 0.3 * (matrix * vec_in);

  EXPECT_EQ(mfem_result.Size(), expr_result.Size());
  for (int i = 0; i < rows; i++) {
    EXPECT_DOUBLE_EQ(mfem_result[i], expr_result[i]);
  }
  MPI_Barrier(MPI_COMM_WORLD);
}

TEST(ExprTemplates, SmallMixedExprSingleAlloc)
{
  MPI_Barrier(MPI_COMM_WORLD);
  constexpr int rows = 10;
  auto [lhs, rhs]    = sample_vectors(rows);

  constexpr int cols    = 12;
  auto [matrix, vec_in] = sample_matvec(rows, cols);

  mfem::Vector mfem_result(rows);
  mfem::Vector expr_result(rows);

  // Scratchpad vectors
  mfem::Vector matvec(rows);
  mfem::Vector vec_negate_scale(rows);

  // -lhs + rhs * 3.0 - 0.3 * (matrix * vec_in)
  matrix.Mult(vec_in, matvec);
  add(-1.0, lhs, 3.0, rhs, vec_negate_scale);
  add(vec_negate_scale, -0.3, matvec, mfem_result);

  evaluate(-lhs + rhs * 3.0 - 0.3 * (matrix * vec_in), expr_result);

  EXPECT_EQ(mfem_result.Size(), expr_result.Size());
  for (int i = 0; i < rows; i++) {
    EXPECT_DOUBLE_EQ(mfem_result[i], expr_result[i]);
  }
  MPI_Barrier(MPI_COMM_WORLD);
}

TEST(ExprTemplates, LargeMixedExpr)
{
  MPI_Barrier(MPI_COMM_WORLD);
  constexpr int rows = 10000;
  auto [lhs, rhs]    = sample_vectors(rows);

  constexpr int cols    = 1200;
  auto [matrix, vec_in] = sample_matvec(rows, cols);

  mfem::Vector mfem_result(rows);
  mfem::Vector expr_result(rows);

  // -lhs + rhs * 3.0 - 0.3 * (matrix * vec_in)
  mfem::Vector matvec(rows);
  matrix.Mult(vec_in, matvec);

  mfem::Vector vec_negate_scale(rows);
  add(-1.0, lhs, 3.0, rhs, vec_negate_scale);
  add(vec_negate_scale, -0.3, matvec, mfem_result);

  expr_result = -lhs + rhs * 3.0 - 0.3 * (matrix * vec_in);

  EXPECT_EQ(mfem_result.Size(), expr_result.Size());
  for (int i = 0; i < rows; i++) {
    EXPECT_DOUBLE_EQ(mfem_result[i], expr_result[i]);
  }
  MPI_Barrier(MPI_COMM_WORLD);
}

TEST(ExprTemplates, ComplexExprLambda)
{
  MPI_Barrier(MPI_COMM_WORLD);
  constexpr int rows = 10;
  auto [lhs, rhs]    = sample_vectors(rows);

  constexpr int cols    = 12;
  auto [matrix, vec_in] = sample_matvec(rows, cols);

  mfem::Vector matvec(rows);
  matrix.Mult(vec_in, matvec);

  mfem::Vector vec_negate_scale(rows);
  add(-1.0, lhs, 3.0, rhs, vec_negate_scale);

  mfem::Vector mfem_result(rows);
  add(vec_negate_scale, -0.3, matvec, mfem_result);

  auto lambda_expr = [](const auto& l, const auto& r, const auto& mat, const auto& vec) {
    return -l + r * 3.0 - 0.3 * (mat * vec);
  };

  mfem::Vector expr_result = lambda_expr(lhs, rhs, matrix, vec_in);

  EXPECT_EQ(mfem_result.Size(), expr_result.Size());
  for (int i = 0; i < rows; i++) {
    EXPECT_DOUBLE_EQ(mfem_result[i], expr_result[i]);
  }
  MPI_Barrier(MPI_COMM_WORLD);
}

int main(int argc, char* argv[])
{
  int result = 0;

  ::testing::InitGoogleTest(&argc, argv);

  MPI_Init(&argc, &argv);

  axom::slic::SimpleLogger logger;

  result = RUN_ALL_TESTS();

  MPI_Finalize();

  return result;
}
